"""This module provide functionality to combine multiple images."""

from typing import List, Tuple

import nibabel as nib
import numpy as np


def overlay_image(image1: np.ndarray, image2: np.ndarray) -> np.ndarray:
    """Overlay image1 with image2.

    Args:
        image1 (np.ndarray): an image
        image2 (np.ndarray): another image

    Raises:
        ValueError: If image shapes are not the same.

    Returns:
        np.ndarray: overlayed image where image2 is on top
    """
    if image1.shape != image2.shape:
        raise ValueError("Shapes of images are not the same.")

    return np.where(np.isclose(image2, 0.), image1, image2)


def overlay_images_with_affine(
        image1: Tuple[np.ndarray, np.ndarray],
        image2: Tuple[np.ndarray,
                      np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
    """Overlay image1 with image 2 if affines are the same.

    Args:
        image1 (Tuple[np.ndarray, np.ndarray]): (image, affine)
        image2 (Tuple[np.ndarray, np.ndarray]): (image, affine)

    Raises:
        ValueError: If affines do not match.

    Returns:
        Tuple[np.ndarray, np.ndarray]: overlayed image and affine
    """
    if not np.allclose(affine1, affine2):
        raise ValueError("Affines are not the same enough for overlay.")

    return overlay_image(image1, image2), affine1


def overlay_multiple_images_with_affines(
    images_with_affine: List[Tuple[np.ndarray, np.ndarray]]
) -> Tuple[np.ndarray, np.ndarray]:
    """Overlay list of images if their affines are the same.

    Args:
        images_with_affine (List[Tuple[np.ndarray, np.ndarray]]): List of images
            in ascending order. First image is background, and last image is 
            foreground.

    Raises:
        ValueError: If length of list is zero.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Overlayed image with affine.
    """
    if len(images_with_affine) == 0:
        raise ValueError("List of images is empty.")
    elif len(images_with_affine) == 1:
        return images_with_affine[0]
    else:
        return overlay_images_with_affine(
            image1=images_with_affine[0],
            image2=overlay_multiple_images_with_affines(
                images_with_affine[1:]))


def sum_images_with_affines(
        image1: Tuple[np.ndarray, np.ndarray],
        image2: Tuple[np.ndarray,
                      np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
    """Sum image1 with image2 if affines are the same.

    Args:
        image1 (Tuple[np.ndarray, np.ndarray]): (image, affine)
        image2 (Tuple[np.ndarray, np.ndarray]): (image, affine)

    Raises:
        ValueError: If affines do not match.

    Returns:
        Tuple[np.ndarray, np.ndarray]: summed image and affine
    """
    if not np.allclose(affine1, affine2):
        raise ValueError("Affines are not the same enough for overlay.")

    return np.add(image1, image2), affine1


def sum_multiple_images_with_affines(
    images_with_affine: List[Tuple[np.ndarray, np.ndarray]]
) -> Tuple[np.ndarray, np.ndarray]:
    """Sum list of images if their affines are the same.

    Args:
        images_with_affine (List[Tuple[np.ndarray, np.ndarray]]): List of images

    Raises:
        ValueError: If length of list is zero.

    Returns:
        Tuple[np.ndarray, np.ndarray]: summed image with affine.
    """
    if len(images_with_affine) == 0:
        raise ValueError("List of images is empty.")
    elif len(images_with_affine) == 1:
        return images_with_affine[0]
    else:
        return sum_images_with_affines(
            image1=images_with_affine[0],
            image2=overlay_multiple_images_with_affines(
                images_with_affine[1:]))