"""The module provides transformations for features (discrete image values)."""

from typing import Dict, Tuple, Union

import numpy as np
from scipy.stats import mode


def translate_features(
    features: np.ndarray, feature_mapping: Dict[int, int]
) -> np.ndarray:
    """
    Translate integer features to new ones.

    Note that, there was a discussion about how to efficiently implement this:
    1.
    2. vectorized with searchsorted and dict as arrays (Factor 6)
    3. vectorized application of dictionaries __getitem__ (Factor 2)
    4. vectorized application of dictionary (Factor 1)

    Parameters
    ----------
    features : np.ndarray
        arrays of features contained in feature_mapping.keys().
    feature_mapping : Dict[int, int]
        dictionary with positive integer keys.

    Returns
    -------
    np.ndarray
        translated features.

    """
    feature_keys = np.array(list(feature_mapping.keys()))
    feature_values = np.array(list(feature_mapping.values()))

    mapping_array = np.zeros(feature_keys.max() + 1,
                             dtype=feature_values.dtype)
    mapping_array[feature_keys] = feature_values
    return mapping_array[features]


def superimpose_features(
    features_a: np.ndarray,
    features_b: np.ndarray,
    feature_superposition: Union[Dict[Tuple[int, int], int], np.ndarray],
) -> np.ndarray:
    """
    Superimpose two feature sets defined by feature_superposition.

    Parameters
    ----------
    features_a : np.ndarray
        arrays of features contained in feature_mapping.keys()[0].
    features_b : np.ndarray
        arrays of features contained in feature_mapping.keys()[1].
    feature_superposition : Union[Dict[Tuple[int, int], int], np.ndarray]
        dictionary with two positive integer keys.

    Returns
    -------
    np.ndarray
        superimposed features.

    """
    if type(feature_superposition) == dict:
        feature_a_keys = np.array(list(feature_superposition.keys()))[:, 0]
        feature_b_keys = np.array(list(feature_superposition.keys()))[:, 1]
        superposition_features = np.array(list(feature_superposition.values()))

        superposition_array = np.zeros(
            (feature_a_keys.max() + 1, feature_b_keys.max() + 1),
            dtype=superposition_features.dtype,
        )
        superposition_array[feature_a_keys,
                            feature_b_keys] = superposition_features
    else:
        superposition_array = feature_superposition

    return superposition_array[features_a, features_b]


def extract_feature_surfaces(
    features: np.ndarray,
) -> np.ndarray:
    """
    Extract feature boundaries of an image.

    Parameters
    ----------
    features : np.ndarray
        discrete feature image.

    Returns
    -------
    surface_labels : np.ndarray
        boundaries of image.

    """
    directions_encoded = np.concatenate(
        [
            np.roll(features, shift=direction, axis=dim)[None]
            for dim in range(len(features.shape))
            for direction in [-1, 1]
        ]
    )
    surfaces = ~np.all(directions_encoded == features, axis=0)
    surface_labels = np.multiply(surfaces, features)

    return surface_labels


def extract_labeled_feature_inner_surfaces(
        features: np.array,
        label: int,
) -> np.array:
    """
    Extract neighbor labeled feature boundaries of a label.

    The labels depend on the surrounding of a boundary voxel.

    Parameters
    ----------
    features : np.array
        discrete feature image.
    label: int
        specify which label to look at.

    Returns
    -------
    surface_labels : np.ndarray
        labeled boundaries of image. Note that boundaries to background will be
        0 if background has label 0.

    """
    if features.dtype != int:
        features = features.astype(int)

    unique_features = np.unique(features)

    if label not in unique_features:
        raise Warning(f"Label {label} does not appear. Empty array returned.")
        return np.empty(features.shape)

    neighbors_encoded = np.concatenate(
        [
            np.roll(features, shift=direction, axis=dim)[None]
            for dim in range(len(features.shape))
            for direction in [-1, 1]
        ]
    )
    surfaces = ~np.all(neighbors_encoded == features, axis=0)
    surfaces[features != label] = 0

    neighbors_encoded = np.searchsorted(unique_features, neighbors_encoded)
    one_hot_encoded_neighbors = np.eye(
        len(unique_features), dtype=int,
    )[neighbors_encoded]

    neighbor_counts = np.sum(one_hot_encoded_neighbors, axis=0)
    neighbor_counts[..., np.where(unique_features == label)[0].item()] = 0
    neighbor_index = np.argmax(neighbor_counts, axis=-1)

    neighbors = unique_features[neighbor_index]
    surface_neighbors = np.multiply(surfaces, neighbors)

    return surface_neighbors


def extract_labeled_feature_outer_surfaces(
        features: np.array,
        label: int,
) -> np.array:
    """
    Extract neighbors of a boundary of a label.

    The labels depend on the surrounding of a boundary voxel.

    Parameters
    ----------
    features : np.array
        discrete feature image.
    label: int
        specify which label to look at.

    Returns
    -------
    surface_labels : np.ndarray
        labeled boundaries of image. Note that boundaries to background will be
        0 if background has label 0.

    """
    if features.dtype != int:
        features = features.astype(int)

    directions_encoded = np.concatenate(
        [
            np.roll(features, shift=direction, axis=dim)[None]
            for dim in range(len(features.shape))
            for direction in [-1, 1]
        ]
    )
    outer_surfaces = ~np.all(directions_encoded == label, axis=0)
    inner_surfaces = np.any(directions_encoded == label, axis=0)

    surfaces = np.multiply(outer_surfaces, inner_surfaces)
    surface_labels = np.multiply(surfaces, features)

    return surface_labels
