#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module provides functionality to deal with DICOM files."""

import datetime as dt
from pathlib import Path
from typing import Tuple

import pydicom


def read_subjects_age_gender(dcm_path: Path) -> Tuple[int, str]:
    """
    Read age and gender for a subject from a DCM file.

    Parameters
    ----------
    dcm_path : Path
        Path to a DCM file.

    Returns
    -------
    Tuple[int, str]
        (age, gender).

    """
    ds = pydicom.dcmread(dcm_path)

    try:
        age = int(ds["PatientAge"].value)
    except AssertionError:
        birth_date = ds["PatientBirthDate"].value
        birth_date = dt.date(year=birth_date[:4],
                             month=birth_date[4:6],
                             day=birth_date[6:8])
        acq_date = ds["AcquisitionDate"].value
        acq_date = dt.date(year=acq_date[:4],
                           month=acq_date[4:6],
                           day=acq_date[6:8])
        age = acq_date.year - birth_date.year
        if acq_date.month < birth_date.month:
            age -= 1
        elif (
                acq_date.month == birth_date.month
        ) and (acq_date.day < birth_date.day):
            age -= 1

    gender = ds["PatientSex"].value

    return age, gender
