#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 19 11:44:28 2024

@author: jsteiglechner
"""

from typing import Tuple

import itk
import numpy as np


def extract_elastix_parameters(
        transform_parameters: itk.ParameterObject,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extract parameters of elastix object to affine parameters.

    Parameters
    ----------
    transform_parameters : itk.ParameterObject
        Parameters of operation.

    Returns
    -------
    center : TYPE
        center of rotation point.
    elx_parameters : TYPE
        transform parameters.

    """
    parameter_map = transform_parameters.GetParameterMap(0)

    center = np.array(
        [float(p) for p in parameter_map['CenterOfRotationPoint']]
    )
    elx_parameters = np.array(
        [float(p) for p in parameter_map['TransformParameters']]
    )

    return center, elx_parameters


def get_transformation_matrix(
        itk_transform: itk.Transform,
) -> np.ndarray:
    """
    Extract relevant a linear itk transform to generate the affine matrix.

    Parameters
    ----------
    itk_transform : itk.Transform
        Linear transform.

    Returns
    -------
    transformation_matrix : TYPE
        Affine transformation matrix.

    """
    affine_direction = itk.array_from_matrix(itk_transform.GetMatrix())
    affine_origin = np.asarray(itk_transform.GetTranslation())

    # Conclude affine information
    transformation_matrix = np.eye(4, dtype=float)
    transformation_matrix[:3, :3] = affine_direction
    transformation_matrix[:3, 3] = affine_origin

    return transformation_matrix


def build_euler_transformation(
        center: np.ndarray,
        transform: np.ndarray,
) -> itk.Euler3DTransform:
    """
    Build an Euler transformation from elastix parameters.

    Parameters
    ----------
    center : np.ndarray
        Center of rotation point in shape (dim, ).
    transform : np.ndarray
        Transform parameters in shape (dim * (dim+1), ).

    Returns
    -------
    euler_transform : TYPE
        Applicable Euler transform.

    """
    euler_transform = itk.Euler3DTransform[itk.D].New()

    fixed_parameters = itk.OptimizerParameters[itk.D](len(center))
    for i, p in enumerate(center):
        fixed_parameters[i] = p
    euler_transform.SetFixedParameters(fixed_parameters)

    itk_parameters = itk.OptimizerParameters[itk.D](len(transform))
    for i, p in enumerate(transform):
        itk_parameters[i] = p
    euler_transform.SetParameters(itk_parameters)

    return euler_transform


def build_affine_transformation(
        center: np.ndarray,
        transform: np.ndarray,
) -> itk.AffineTransform:
    """
    Build an Affine transformation from elastix parameters.

    Parameters
    ----------
    center : np.ndarray
        Center of rotation point in shape (dim, ).
    transform : np.ndarray
        Transform parameters in shape (dim * (dim+1), ).

    Returns
    -------
    affine_transform : TYPE
        Applicable affine transform.

    """
    affine_transform = itk.AffineTransform[itk.D, len(center)].New()

    fixed_parameters = itk.OptimizerParameters[itk.D](len(center))
    for i, p in enumerate(center):
        fixed_parameters[i] = p
    affine_transform.SetFixedParameters(fixed_parameters)

    itk_parameters = itk.OptimizerParameters[itk.D](len(transform))
    for i, p in enumerate(transform):
        itk_parameters[i] = p
    affine_transform.SetParameters(itk_parameters)

    return affine_transform
