#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module provides functionality for handling NIfTI images with nibabel.

Created on Fri Aug  4 14:19:10 2023

@author: jsteiglechner
"""

from pathlib import Path
from typing import Tuple, Union

import nibabel as nib
import numpy as np


def read_nii(filepath: Union[str, Path]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Read brain MRI and corresponding affine from .nii file.

    Note that nibabel generates a fortran ordered array.

    Parameters
    ----------
    filepath : Union[str, Path]
        Path to MRI.

    Returns
    -------
    image : np.ndarray
        Image data.
    affine_matrix : np.ndarray
        Corresponsing affine matrix.

    """
    nifti = nib.load(filepath)
    image = nifti.dataobj[:]
    affine_matrix = nifti.affine

    return image, affine_matrix


def save_nii(
        image: np.ndarray,
        affine_matrix: np.ndarray,
        filepath: Union[str, Path],
        dtype: np.dtype = float,
) -> None:
    """
    Save brain MRI with affine matrix in NIfTI file format.

    Parameters
    ----------
    image : np.ndarray
        Image data.
    affine_matrix : np.ndarray
        Corresponding affine matrix.
    filepath : Union[str, Path]
        Path where MRI should be saved.
    dtype : np.dtype, optional
        The type of the saved array. The default is float.

    Returns
    -------
    None
        DESCRIPTION.

    """
    nifti = nib.Nifti1Image(image, affine_matrix, dtype=dtype)
    nib.save(nifti, filepath)


def get_datatype(image: np.ndarray) -> np.dtype:
    """
    Get datatype of image data.

    Parameters
    ----------
    image : np.ndarray
        NIfTI image dataobj.

    Returns
    -------
    datatype : np.dtype
        numpy datatype.

    """
    if np.all(np.equal(np.mod(image, 1), 0)):
        if np.min(image) >= 0:
            if np.max(image) <= (2**8 - 1):
                datatype = np.uint8
            elif np.max(image) <= (2**16 - 1):
                datatype = np.uint16
            else:
                datatype = np.uint32
        elif (np.min(image) > -2**15) and (np.max(image) < 2**15):
            datatype = np.int16
        else:
            datatype = np.int32
    else:
        datatype = np.dtype('float')

    return datatype


def convert_to_original_datatype(
        image: np.ndarray,
        datatype: np.dtype,
) -> np.ndarray:
    """
    Convert an image to a certain original datatype.

    Parameters
    ----------
    image : np.ndarray
        Image dataobject.
    datatype : np.dtye
        original datatype.

    Returns
    -------
    image : np.ndarray
        Converted image.

    """
    if datatype == np.dtype('float'):
        image = image.astype(float)
    elif datatype == np.uint8:
        image = nib.casting.float_to_int(image, 'uint8')
    elif datatype == np.uint16:
        image = nib.casting.float_to_int(image, 'uint16')
    elif datatype == np.uint32:
        image = nib.casting.float_to_int(image, 'uint32')
    elif datatype == np.int16:
        image = nib.casting.float_to_int(image, 'int16')
    else:
        image = nib.casting.float_to_int(image, 'int32')

    return image
