"""Module provides functions for feature scaling.

That contains intensity normalization and standarization.

For further implementation, see
* https://en.wikipedia.org/wiki/Whitening_transformation
* histogram based normalization
"""

from typing import Union

import numpy as np


def standardize(feature: np.ndarray) -> np.ndarray:
    """Standardize feature to zero mean and unit variance."""
    feature = feature.astype(float)

    intensity_mean = np.mean(feature)
    intensity_std = np.std(feature)

    if intensity_std > 0:
        return (feature - intensity_mean) / intensity_std
    else:
        return feature * 0.


def rescale_min_max(feature: np.ndarray) -> np.ndarray:
    """Rescale the range of features to scale the range [0, 1] or [-1, 1]."""
    return (feature - np.min(feature)) / (np.max(feature) - np.min(feature))


def rescale_range(feature: np.ndarray, minimum: float,
                  maximum: float) -> np.ndarray:
    """Rescale a range between arbitrary set of values [minimum, maximum]."""
    return minimum + rescale_min_max(feature) * (maximum - minimum)


def normalize_mean(feature: np.ndarray) -> np.ndarray:
    """Rescale the difference to mean to scale [minimum, maximum]."""
    return (feature - np.mean(feature)) / (np.max(feature) - np.min(feature))


def scale_to_unit(feature: np.ndarray, norm='L2') -> np.ndarray:
    """Scale the components of a feature such that its vector has length one."""
    if norm == 'L1':
        return feature / np.sum(feature)
    elif norm == 'L2':
        return feature / np.norm(feature)
    else:
        raise ValueError('The provided norm is not supported.')


def equalize_histogram(feature: np.ndarray,
                       number_bins: int = None) -> np.ndarray:
    """Adjust the contrast using the feature's histogram.

    Note:
        Be careful if feature is integer.
    """
    if number_bins is None:
        feature_range = np.max(feature) - np.min(feature)
        step_size = 10**(np.around(np.log10(feature_range)) - 3)
        number_bins = np.ceil(feature_range / step_size).astype(np.int)

    hist, grey_levels = np.histogram(feature.flatten(),
                                     number_bins,
                                     density=True)

    cdf = np.cumsum(hist)  # cumulative distribution function

    # normalize
    # cdf = np.max(grey_levels)

    feature_equalized = np.interp(feature.flatten(), grey_levels[:-1], cdf)

    return feature_equalized.reshape(feature.shape)


def match_histogram(feature: np.ndarray,
                    reference: np.ndarray,
                    feature_mask: np.ndarray = None,
                    reference_mask: np.ndarray = None,
                    number_bins: int = 500,
                    hmin: Union[int, float] = -99999,
                    hmax: Union[int, float] = 99999) -> np.ndarray:
    """Match Histogram from feature to reference.

    Args:
        feature (np.ndarray): [description]
        reference (np.ndarray): [description]
        feature_mask (np.ndarray, optional): [description]. Defaults to None.
        reference_mask (np.ndarray, optional): [description]. Defaults to None.
        number_bins (int, optional): [description]. Defaults to 500.
        hmin (Union[int, float], optional): [description]. Defaults to -99999.
        hmax (Union[int, float], optional): [description]. Defaults to 99999.

    Returns:
        np.ndarray: [description]
    """

    if feature_mask is None:
        feature_mask = ~(np.abs(feature) < 1e-6)
    if reference_mask is None:
        reference_mask = ~(np.abs(reference) < 1e-6)

    feature = np.maximum(np.minimum(feature, hmax), hmin)
    reference = np.maximum(np.minimum(reference, hmax), hmin)

    feature_hist, feature_grey_levels = np.histogram(
        feature[feature_mask],
        number_bins,
        density=False,
    )
    feature_hist = feature_hist / np.sum(feature_mask)

    reference_hist, reference_grey_levels = np.histogram(
        reference[reference_mask],
        number_bins,
        density=False,
    )
    reference_hist = reference_hist / np.sum(reference_mask)
    reference_bin_levels = (reference_grey_levels[:-1] +
                            reference_grey_levels[1:]) / 2

    feature_cdf = np.cumsum(feature_hist)
    reference_cdf = np.cumsum(reference_hist)

    interp_values = np.interp(feature_cdf, reference_cdf, reference_bin_levels)

    feature_indices = np.digitize(
        feature, feature_grey_levels[:-1], right=False) - 1

    feature_matched = np.where(feature_mask, interp_values[feature_indices], 0)

    return feature_matched


def quantile_clipping(
        feature: np.array,
        lower_quantile: float = 0.01,
        upper_quantile: float = 0.99,
) -> np.ndarray:
    lower_limit = np.quantile(feature, lower_quantile)
    upper_limit = np.quantile(feature, upper_quantile)

    feature = np.where(
        feature < lower_limit,
        np.min(feature[feature >= lower_limit]),
        feature,
    )
    feature = np.where(
        feature > upper_limit,
        np.max(feature[feature <= upper_limit]),
        feature,
    )

    return feature
