"""
Provide methods to denoise mp2rage MRI.

adapted from RobustCombination function from Jose Marques,
https://github.com/JosePMarques/MP2RAGE-related-scripts

Applications:
https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0099676
https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0210803
"""

from pathlib import Path
from typing import Union

import nibabel as nib
import numpy as np

from .unite_mp2rage import unite_mp2rage_robust


def quadratic_formula_positive(a, b, c):
    """Give positive solution of quadratic formula."""
    return (-b + np.sqrt(b**2 - 4 * a * c)) / (2 * a)


def quadratic_formula_negative(a, b, c):
    """Give negative solution of quadratic formula."""
    return (-b - np.sqrt(b**2 - 4 * a * c)) / (2 * a)


def combine_mp2rage_robust(
        uniform_filepath: Union[str, Path],
        inv1_fielpath: Union[str, Path],
        inv2_filepath: Union[str, Path],
        denoised_output_filepath: Union[str, Path],
        regularization_factor: float = 6,
        value_mode: str = "complex",
) -> None:
    """
    Suppress background noise inside and outside the skull.

    Note
    ----
        usually the regularization factor shouldn't be greater then 10, but
        that is not the case when the image is bias field corrected, in which
        case the noise estimated at the edge of the image might not be such a
        good measure.

    Parameters
    ----------
    uniform_filepath : Union[str, Path]
        DESCRIPTION.
    inv1_fielpath : Union[str, Path]
        DESCRIPTION.
    inv2_filepath : Union[str, Path]
        DESCRIPTION.
    denoised_output_filepath : Union[str, Path]
        DESCRIPTION.
    regularization_factor : float, optional
        DESCRIPTION. The default is 6.
    value_mode : str, optional
        DESCRIPTION. The default is 'absolute'.

    Returns
    -------
    None
        DESCRIPTION.

    """
    if value_mode not in ["absolute", "complex"]:
        raise ValueError("Value mode has to be absolute or complex.")

    mp2rage_nii = nib.load(uniform_filepath)
    inv1_nii = nib.load(inv1_fielpath)
    inv2_nii = nib.load(inv2_filepath)

    mp2rage_img = mp2rage_nii.dataobj[:]
    if value_mode == "complex":
        inv1_img = inv1_nii.dataobj[:]
        inv2_img = inv2_nii.dataobj[:]
    elif value_mode == "absolute":
        inv1_img = np.abs(inv1_nii.dataobj[:])
        inv2_img = np.abs(inv2_nii.dataobj[:])

    # rescale the range of features to scale the range[-.5, .5]
    if mp2rage_img.min() >= 0 and mp2rage_img.max() >= 0.51:
        # assumes that it is getting only positive values
        mp2rage_img = (mp2rage_img - mp2rage_img.max() / 2) / mp2rage_img.max()
        integerformat = False
    elif mp2rage_img.min() < 0:
        # rescale min max
        min_mp2rage = np.min(mp2rage_img)
        max_mp2rage = np.max(mp2rage_img)
        mp2rage_img = (mp2rage_img - min_mp2rage) / (max_mp2rage -
                                                     min_mp2rage) - .5
        inv1_img = inv1_img / (max_mp2rage - min_mp2rage) * 2
        inv2_img = inv2_img / (max_mp2rage - min_mp2rage) * 2
        integerformat = False
    else:
        integerformat = True

    # computes correct polarity to inv1 image
    inv1_img = np.sign(mp2rage_img) * inv1_img

    # because the MP2RAGE INV1 and INV2 is a sum of squares data, while the
    # MP2RAGEimg is a phase sensitive coil combination, some more maths have to
    # be performed to get a better INV1 estimate which here is done by assuming
    # both INV2 is closer to a real phase sensitive combination
    inv1_positive = quadratic_formula_positive(
        -mp2rage_img,
        inv2_img,
        -inv2_img**2 * mp2rage_img,
    )
    inv1_negative = quadratic_formula_negative(
        -mp2rage_img,
        inv2_img,
        -inv2_img**2 * mp2rage_img,
    )

    inv1_estimated = inv1_img.copy()

    inv1_estimated[np.absolute(inv1_img - inv1_positive) > np.absolute(
        inv1_img - inv1_negative)] = inv1_negative[np.absolute(
            inv1_img - inv1_positive) > np.absolute(inv1_img - inv1_negative)]
    inv1_estimated[np.absolute(inv1_img - inv1_positive) <= np.absolute(
        inv1_img - inv1_negative)] = inv1_positive[np.absolute(
            inv1_img - inv1_positive) <= np.absolute(inv1_img - inv1_negative)]

    # Apply regularization of noise
    noiselevel = regularization_factor * np.mean(inv2_img[:, -11:, -11:])
    mp2rage_robust = unite_mp2rage_robust(inv1_estimated, inv2_img,
                                          noiselevel**2)

    if integerformat:
        denoised_mp2rage_img = mp2rage_robust
    else:
        denoised_mp2rage_img = np.round(4095 * (mp2rage_robust + 0.5))
        denoised_mp2rage_img = nib.casting.float_to_int(
            denoised_mp2rage_img, 'int16')

    # save image
    denoised_mp2rage_nii = nib.Nifti1Image(denoised_mp2rage_img,
                                           mp2rage_nii.affine,
                                           mp2rage_nii.header)
    nib.save(denoised_mp2rage_nii, denoised_output_filepath)
