"""This module provides methods to unify mp2rage MRI."""

import numpy as np


def unite_mp2rage_robust(
        TI1_img: np.ndarray,
        TI2_img: np.ndarray,
        beta: float = 1e-5,
) -> np.ndarray:
    """Combine the two images in a robust way.

    See also:
        https://archive.ismrm.org/2013/0269.html
        https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0099676#pone.0099676.e004
    """
    return (np.conj(TI1_img) * TI2_img - beta) / (TI1_img**2 + TI2_img**2 +
                                                  2 * beta)


def unite_mp2rage(TI1_img: np.ndarray,
                  TI2_img: np.ndarray,
                  beta: float = 1e-5) -> np.ndarray:
    """Combine the two images acquired at TI1 and TI2."""
    sum_imgs = np.square(TI1_img) + np.square(TI2_img)

    mp2rage = np.real(np.conj(TI1_img) * TI2_img) / np.clip(
        sum_imgs, a_min=beta, a_max=np.max(sum_imgs))

    return mp2rage
