#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module performs affine transformations on basis of numpy.

Created on Thu Jan 16 11:48:08 2025

@author: jsteiglechner
"""

from typing import Tuple

import numpy as np

from .interpolations import (
    get_domain_values_at_positions,
    linear_distance_weighting_of_points,
    linear_interpolation,
    nearest_neightbour,
    weighted_majority_resampling,
)
from .complete_affine_transform import complete_target_affine_space


def get_relative_kernel_coordinates(kernel_size: int) -> np.ndarray:
    """Build the relative kernel coordinates from origin."""
    if kernel_size == 1:
        return np.array([[0, 0, 0]], dtype=np.int8).T
    elif kernel_size == 2:
        return np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1],
                         [1, 0, 0], [1, 0, 1], [1, 1, 0], [1, 1, 1]],
                        dtype=np.int8).T  # yapf: disable
    elif kernel_size == 3:
        return np.array(
            [[-1, -1, -1], [-1, -1, 0], [-1, -1, 1], [-1, 0, -1], [-1, 0, 0],
             [-1, 0, 1], [-1, 1, -1], [-1, 1, 0], [-1, 1, 1],
             [0, -1, -1], [0, -1, 0], [0, -1, 1], [0, 0, -1], [0, 0, 0],
             [0, 0, 1], [0, 1, -1], [0, 1, 0], [0, 1, 1],
             [1, -1, -1], [1, -1, 0], [1, -1, 1], [1, 0, -1], [1, 0, 0],
             [1, 0, 1], [1, 1, -1], [1, 1, 0], [1, 1, 1]],
            dtype=np.int8).T  # yapf: disable
    else:
        raise ValueError(f"Kernel size {kernel_size} isn't implemented yet.")


def kernel_coordinates(set_of_points: np.ndarray,
                       kernel_size: int) -> np.ndarray:
    """Build the associated points in a kernel."""
    floored_set_of_points = np.floor(set_of_points).astype(np.int16)
    relative_kernel_coordinates = get_relative_kernel_coordinates(
        kernel_size=kernel_size)

    absolute_kernel_coordinates = np.add(
        floored_set_of_points[..., np.newaxis],
        relative_kernel_coordinates).astype(np.int16)

    if np.max(absolute_kernel_coordinates) < 128:
        absolute_kernel_coordinates = absolute_kernel_coordinates.astype(
            np.int8)
    else:
        absolute_kernel_coordinates = absolute_kernel_coordinates.astype(
            np.int16)

    return absolute_kernel_coordinates


def get_raster_augmented_coordinates(
        raster_shape: Tuple[int, int, int]) -> np.ndarray:
    """Extract all augmented indices in a raster."""
    raster_coordinates = np.ones((4, ) + raster_shape, dtype=np.int16)
    raster_coordinates[:3] = np.indices(raster_shape)

    return raster_coordinates


def affine_coordinates_transform(affine_map: np.ndarray,
                                 set_of_points: np.ndarray) -> np.ndarray:
    """Apply an affine map on a set of augmented points."""
    x_set = np.moveaxis(set_of_points, 0, -2)
    y_set = np.matmul(affine_map, x_set, dtype=np.float16)
    y_set = np.moveaxis(y_set, -2, -1)

    return y_set[..., :3]


def interpolate_image(
        image: np.ndarray,
        affine_transform_matrix: np.ndarray,
        target_shape: np.ndarray,
        interpolation: str,
        interpolation_kernel_size: int = 2,
) -> np.ndarray:
    """
    Relocate pixels requiring intensity interpolation.

    Approximate the value of moved pixels.

    """
    codomain_raster_coordinates = get_raster_augmented_coordinates(
        raster_shape=target_shape)
    domain_coordinates = affine_coordinates_transform(
        affine_map=affine_transform_matrix,
        set_of_points=codomain_raster_coordinates)
    del codomain_raster_coordinates
    domain_kernel_coordinates = kernel_coordinates(
        set_of_points=domain_coordinates,
        kernel_size=interpolation_kernel_size)
    domain_values = get_domain_values_at_positions(
        domain_values=image, set_of_kernel_points=domain_kernel_coordinates)
    domain_weights = linear_distance_weighting_of_points(
        set_of_points=domain_coordinates,
        set_of_kernel_points=domain_kernel_coordinates)
    del domain_kernel_coordinates

    if interpolation == 'linear':
        del domain_coordinates
        return linear_interpolation(weights=domain_weights,
                                    values=domain_values)
    elif interpolation == 'nearest':
        return nearest_neightbour(values=image,
                                  set_of_points=domain_coordinates)
    elif interpolation == 'majority_voted':
        del domain_coordinates
        return weighted_majority_resampling(
            weights=domain_weights.astype(np.float16),
            values=domain_values,
            unique_values=np.unique(image),)
    else:
        raise ValueError(
            f'Interpolation method {interpolation} is not available.')


def affine_transform_image(
        image: np.ndarray,
        affine_matrix: np.ndarray,
        target_affine_matrix: np.ndarray,
        target_shape: Tuple[int, int, int] = None,
        interpolation: str = 'linear',
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Perform an affine transformation to an image.

    Parameters
    ----------
    image : np.ndarray
        Domain image.
    affine_matrix : np.ndarray
        Domain affine space.
    target_affine_matrix : np.ndarray
        Codomain affine space.
    target_shape : Tuple[int, int, int], optional
        Codomain image shape. The default is None.
    interpolation : str, optional
        Interpolation method. The default is 'linear'.

    Raises
    ------
    ValueError
        If affine spaces contain infinite numbers.
        If image contains infinite numbers.

    Returns
    -------
    np.ndarray
        Codomain image.
    np.ndarray
        Codomain affine.

    """
    if np.any(np.isnan(affine_matrix)) or np.any(
            np.isnan(target_affine_matrix)):
        raise ValueError('Affine Matrices contain infinite numbers.')
    if np.any(np.isnan(image)):
        raise ValueError('Image contains infinite numbers.')

    # Check if transform is necessary
    if target_affine_matrix.shape != (4, 4):
        affine_to_compare = affine_matrix[:3, :3]
    else:
        affine_to_compare = affine_matrix
    if np.allclose(affine_to_compare,
                   target_affine_matrix) and image.shape is target_shape:
        return image, affine_matrix

    target_affine_matrix, target_shape = complete_target_affine_space(
        affine_matrix=affine_matrix,
        target_affine_matrix=target_affine_matrix,
        image_shape=image.shape,
        target_shape=target_shape,
    )

    affine_transform_matrix = np.dot(np.linalg.inv(affine_matrix),
                                     target_affine_matrix)

    transformed_image = interpolate_image(
        image=image,
        affine_transform_matrix=affine_transform_matrix,
        target_shape=target_shape,
        interpolation=interpolation)

    return transformed_image, target_affine_matrix
